# PHP pledge

This is a PHP extension that adds support for OpenBSD's pledge and unveil system calls.

## Requirements

This extension works with >= 7.2 and needs at least OpenBSD 6.4.

## The theory

The [pledge(2) system call](http://man.openbsd.org/OpenBSD-current/man2/pledge.2) allows a program to restrict the types
of operations the program can do after that point. Unlike other similar systems, pledge is specifically designed for
programs that need to use a wide variety of operations on initialisation, but a fewer number after initialisation (when
user input will be accepted).

The pledge system call is supported on OpenBSD >= 5.9

The first call to [unveil(2) system call](http://man.openbsd.org/OpenBSD-current/man2/unveil.2) restricts the filesystem
view. Subsequent calls can open it more. To prevent further unveiling, call unveil with no parameters or drop the unveil
pledge if the program is pledged.

The unveil system call is supported on OpenBSD >= 6.4

## OpenBSD installation

This package is present in OpenBSD packages. To install it, run:

```
pkg_add pecl82-pledge-2.1.1
```

This package can also be installed from ports. To install it, make sure you [have the ports tree](https://www.openbsd.org/faq/ports/ports.html#PortsFetch)
on your machine. Next, install the port:

```
cd /usr/ports/www/pecl-pledge/
env FLAVOR=php82 make FETCH_PACKAGES= install
```

Using the extension:

```
echo 'extension=pledge' > /etc/php-8.2/pledge.ini
php-8.2 -m | grep pledge
```

You can set a pledge and unveil default with ini values:

- openbsd.pledge_promises
- openbsd.pledge_execpromises
- openbsd.unveil

Promises are the pledge promises for your main process, execpromises are for child processes (or the same as the parent
if not specified). The openbsd.unveil ini setting can receive many unveil calls eg:
/html/public:r,/var/www/html/tmp:rw would make the /var/www/html/public directory readable and the
/var/www/html/tmp readable and writeable.

## FPM configuration

You can set promises and unveils in your PHP-FPM config.

```
php_admin_value[openbsd.promises] = stdio rpath wpath cpath fattr flock unveil inet
php_admin_value[openbsd.unveil] = /:r,/htdocs/var/log:rwc,/htdocs/var/cache:rwc
```

An simplified httpd example config:

```
chroot "/var/www"

server "example.com" {
    listen on * port 80
    root "/htdocs/public"
    directory index "index.php"

    # Assets not served by PHP
    location match "\.(css|gif|jpg|png|js)$" {
        pass
    }

    location match "/specific-path-1" {
        request rewrite "/index.php/%1"
        fastcgi socket "/run/php-fpm-specific-path-1.sock"
    }

    location match "/specific-path-2" {
        request rewrite "/index.php/%1"
        fastcgi socket "/run/php-fpm-specific-path-2.sock"
    }

    # The default PHP handler
    location match "^/(.+)$" {
        request rewrite "/index.php/%1"
        fastcgi socket "/run/php-fpm.sock"
    }
}
```

With a simplified PHP-FPM config:

```
[global]
include=/etc/php-fpm.d/*.conf


[specific-path-1]
user = www
group = www
listen.owner = www
listen.group = www
listen.mode = 0660

pm = dynamic
pm.max_children = 5
pm.start_servers = 2
pm.min_spare_servers = 1
pm.max_spare_servers = 3

chroot = /var/www
pm.max_requests = 1

listen = /var/www/run/php-fpm-specific-path-1.sock
php_admin_value[openbsd.promises] = stdio rpath wpath cpath fattr flock unveil
php_admin_value[openbsd.unveil] = /:r,/htdocs/var/log:rwc,/htdocs/var/cache:rwc

[specific-path-2]
user = www
group = www
listen.owner = www
listen.group = www
listen.mode = 0660

pm = dynamic
pm.max_children = 5
pm.start_servers = 2
pm.min_spare_servers = 1
pm.max_spare_servers = 3

chroot = /var/www
pm.max_requests = 1

listen = /var/www/run/php-fpm-specific-path-2.sock
php_admin_value[openbsd.promises] = stdio rpath wpath cpath fattr flock unveil
php_admin_value[openbsd.unveil] = /:r,/htdocs/var/log:rwc,/htdocs/var/cache:rwc

[www]
user = www
group = www
listen.owner = www
listen.group = www
listen.mode = 0660

pm = dynamic
pm.max_children = 5
pm.start_servers = 2
pm.min_spare_servers = 1
pm.max_spare_servers = 3

chroot = /var/www
pm.max_requests = 1

listen = /var/www/run/php-fpm.sock
php_admin_value[openbsd.promises] = stdio rpath wpath cpath fattr flock unveil inet
php_admin_value[openbsd.unveil] = /:r,/htdocs/var/log:rwc,/htdocs/var/cache:rwc
```

Don't forget to call `unveil(null, null);` in your PHP userland to disallow future unveil calls.
See [ctors/pledge-symfony-routing](https://packagist.org/packages/ctors/pledge-symfony-routing) for
Symfony framework support.

## Build

Install a release with PECL (see [security/pledge](https://pecl.php.net/package/pledge)) or build the latest from source:

```
git clone https://github.com/tvlooy/php-pledge
cd php-pledge
phpize
./configure
make
doas make install
```

Add the following to your configuration to enable the extension:

```
extension=pledge
```

Run the tests with:

```
NO_INTERACTION=1 make test
```

## Pledge usage

Note: You will need to include the ```stdio``` promise every time because PHP will not work well without it.

All promises are documented in [the OpenBSD pledge(2) manual page](http://man.openbsd.org/OpenBSD-current/man2/pledge.2).

When pledging your program keep an eye on ```/var/log/messages``` to see your violations.

To allow read/write filesystem access, but not network access:

```php
pledge('stdio rpath wpath cpath');
```

To allow inet/unix socket access and DNS queries, but not filesystem access:

```php
pledge('stdio inet unix dns');
```

If the PHP ```pledge()``` call fails, it will throw a ```\PledgeException```.

The ```error``` pledge will make the call fail but will not kill the program. An example:

```
$ php -dopenbsd.promises='stdio dns' -r 'echo gethostbyname("openbsd.org");'
199.185.178.80

$ php -dopenbsd.promises='stdio error' -r 'echo gethostbyname("openbsd.org");'
openbsd.org

$ php -dopenbsd.promises='stdio' -r 'echo gethostbyname("openbsd.org");'
Abort trap (core dumped)
```

To set promises for an execve child:

```php
pledge(null, 'stdio rpath tty error');
pcntl_exec('/bin/ls');
```

## Unveil usage

```php
# Works
print_r(scandir('/var'));

# Limit read only access to /var/spool
unveil('/var/spool', 'r');

# No longer works
print_r(scandir('/var'));
# Works
print_r(scandir('/var/spool'));

# Disallow more unveils
unveil();

# No longer works
unveil('/var', 'r');
```

If the PHP ```unveil()``` call fails, it will throw a ```\UnveilException```.

## Usage tips

Be careful what you pledge! If you run PHP with mod_php, you will be pledging an Apache child processes! If you pledge
php_fpm you will be pledging it for the lifetime of the process, not just the request!

For your web SAPI, make sure you limit the amount of loaded extensions. If you only need ```phar```, ```pcntl```, ...
in your CLI, then don't load them in your web SAPI.

Unveil can for example be applied to FPM processes in addition to or as an alternative to chroot. Note that PHP-FPM runs
chroot-ed by default on OpenBSD, but you can't pledge chroot, so you can't start FPM with a openbsd.promises.

Use openbsd.execpromises to start your parent process chrooted and apply promises to the worker processes instead, and/or
apply pledge() in your userland code instead.

Pledging CLI processes is the most convenient usecase. Examples:

If you are running the php interactive shell with ```php -a``` you need these promises:

```php
pledge('stdio rpath wpath tty');
```

If you want to pledge Drupal8 or Symfony2 running on FPM, you need at least:

```php
pledge('stdio rpath wpath cpath inet dns flock fattr');
```

Just to serve a ```phpinfo()``` or "hello world" from FPM you need:

```php
pledge('stdio rpath flock inet');
```

You can then further limit filesystem access with unveil(). If you are in web SAPI, remember that you are not limiting
filesystem access for one request but for all subsequent requests. Avoid having to add the ```unveil``` promise by
checking if the process is already running with a restricted view. Eg:

```php
if (is_dir('/etc')) {
    unveil(__DIR__, 'r');
}
pledge('stdio rpath flock inet');
```

## A word about FFI

Note that with PHP >= 7.4 you could also use the Foreign Function Interface (FFI) extension that is in core to call libc
functions. But the FFI extension is not built by default in the OpenBSD port/package. And, it does not come with ini
settings like this pledge extension, so you cannot hook pledge() and unveil() into the PHP engine startup.

```bash
$ cat test_ffi.php       
<?php

$libc = FFI::cdef('
    int unveil(const char *path, const char *permissions);
', 'libc.so.97');

var_dump(count(scandir('/etc')));
$libc->unveil(__DIR__, 'r');
scandir('/etc');

$ php -dextension=ffi test_ffi.php      
int(77)

Warning: scandir(/etc): failed to open dir: No such file or directory in /home/tvl/test_ffi.php on line 9

Warning: scandir(): (errno 2): No such file or directory in /home/tvl/test_ffi.php on line 9
```

### Limiting network calls

There currently is nothing to support "unveiling" of network resources. But this can be achieved with PF. Eg:

```
block out proto {tcp udp} user your_fpm_user
pass out proto tcp to $mysql_db port 3306 user your_fpm_user
pass out proto tcp to $some_rest_api port 443 user your_fpm_user
```
